
import os, sys
import cherrypy
import jinja2
import json
from datetime import datetime

from kizano import getLogger, getConfig, Config
log = getLogger('smartmetertx.server')

from smartmetertx.utils import getMongoConnection
from smartmetertx.controller import SmartMeterController

DEFAULT_UI_PATH = os.path.join( sys.exec_prefix, 'share', 'smartmetertx' )

class MeterServer(SmartMeterController):
    mongo = None

    def __init__(self, config: Config):
        super(MeterServer, self)
        self.config = config
        self.db = getMongoConnection(config).get_database(config['mongo'].get('dbname', 'smartmetertx'))

    def __del__(self):
        self.close()

    def close(self):
        if self.mongo:
            self.mongo.close()
            self.mongo = None

    @cherrypy.expose
    def index(self):
        '''
        Home page!
        '''
        return self.returnValue(True, {'hello': 'world'})

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def meterRead(self, date: str = None):
        '''
        Return a meter read for a specified date.
        Gets the full meter read from the DB.
        '''
        result = {}
        if date is None:
            return self.returnValue(False, 'No date specified.')
        try:
            import dateparser
            queryDate = dateparser.parse(date)
            timerange = {
                '$gte': queryDate.replace( hour=max(0, queryDate.hour-1) ),
                '$lt': queryDate.replace( hour=min(23, queryDate.hour+1) )
            }
            result = self.db.meterReads.find_one({'datetime': timerange })
            if result is None:
                return self.returnValue(False, f'No meter read found for {date}')
            log.debug(result)
            del result['_id']
            result['datetime'] = result['datetime'].strftime('%F/%R:%S')
            return self.returnValue(True, result)
        except Exception as e:
            import traceback as tb
            log.error(f'Error getting meter read for {date}: {e}')
            log.error(tb.format_exc())
            return self.returnValue(False, 'uhm, well, this is embarassing :S')

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def meterReads(self, fdate: str = None, tdate: str = None):
        '''
        Return a list of meter reads for a specified date range.
        Gets only the list of values paired with the date as an object/key-value pairing.
        '''
        result = []
        if fdate is None:
            return self.returnValue(False, 'No From Date Specified. Need `fdate`.')
        if tdate is None:
            return self.returnValue(False, 'No To Date Specified. Need `tdate`.')
        import dateparser
        fromDate = dateparser.parse(fdate)
        toDate = dateparser.parse(tdate)
        timerange = {
            '$gte': fromDate,
            '$lt': toDate
        }
        projection = { '_id': False, 'reading': True, 'datetime': True}
        reads = list( self.db.meterReads.find({'datetime': timerange }, projection) )
        for mRead in reads:
            sdate = mRead['datetime'].strftime('%F')
            result.append( [sdate, mRead['reading'] ] )
        return self.returnValue(True, result)

    @cherrypy.expose
    def shutdown(self):
        '''
        Shutdown the server.
        '''
        log.info('Server shutting down...')
        cherrypy.engine.exit()
        log.info('CherryPy Server exit.')

class GoogleGraphsFS(SmartMeterController):
    def __init__(self, uiPath: str = None):
        log.info(f'Serving files from {uiPath}')
        fsloader = jinja2.FileSystemLoader( uiPath )
        self.view = jinja2.Environment(loader=fsloader)
        cherrypy.response.headers['Access-Control-Allow-Origin'] = '*'
        cherrypy.response.headers['Access-Control-Allow-Headers'] = 'Content-Type, Authorization'

    @cherrypy.expose
    def index(self, **kwargs):
        return self.view.get_template('index.html').render(
            page='<p>Index Page.</p>',
            navigation='<li><a href="/user/login">Login</a></li>'
        )

def main():
    '''
    Main application/API entry point.
    '''
    cherrypy._cplogging.LogManager.time = lambda self: datetime.now().strftime('%F %T')
    # Default access_log_format '{h} {l} {u} {t} "{r}" {s} {b} "{f}" "{a}"'
    # h - remote.ip, l - "-", u - login (or "-"), t - time, r - request line, s - status, b - content length
    # f - referer, a - User Agent, o - Host or -, i - request.unique_id, z - UtcTime
    cherrypy._cplogging.LogManager.access_log_format = '{' + json.dumps({
        'time': '{t}',
        'from': '{h}',
        'user': '{u}',
        'host': '{o}',
        'status': '{s}',
        'bytes': '{b}',
        'referer': '{f}',
        'agent': '{a}'
    }) + '}'
    #'{t} from={h} user={u} host={o} status={s} bytes={b} referer="{f}" agent="{a}"'
    config = getConfig()
    ui_path = os.path.realpath( config.get('server', {}).get('ui.path', DEFAULT_UI_PATH) )
    if not os.path.exists(ui_path):
        possible_paths = [
            os.path.join(os.getenv('HOME', '/home/markizano'), '.local', 'share', 'smartmetertx'),
            os.path.join(sys.prefix, 'local', 'share', 'smartmetertx'),
            os.path.join(sys.prefix, 'share', 'smartmetertx'),
        ]
        for ppath in possible_paths:
            if os.path.exists(ppath):
                ui_path = ppath
                break
            else:
                log.debug(f'Could not find UI in path: {ppath}')
        if not os.path.exists(ui_path):
            log.error(f'Could not find UI path: {ui_path}')
            return 1
    apiConfig = {
        'tools.trailing_slash.on': False,
        'tools.json_in.on': True,
        'tools.staticdir.on': False,
    }
    serverConfig = {
        'tools.trailing_slash.on': False,
        'tools.staticdir.on': True,
        'tools.staticdir.dir': ui_path
    }
    log.info('Starting SmartMeterTX Server...')
    log.debug(config)
    smtx = MeterServer(config)
    content = GoogleGraphsFS( serverConfig['tools.staticdir.dir'] )
    log.debug(f'Got config: {config}')
    cherrypy.config.update(config.get('daemon', {}).get('cherrypy', {}))
    cherrypy.tree.mount(smtx, '/api', { '/': apiConfig } )
    cherrypy.tree.mount(content, '/', {'/': serverConfig })
    cherrypy.engine.start()
    cherrypy.engine.block()
    return 0




'''
\x1b[31mModule-Level Documentation!\x1b[0m
'''
import os, sys
import yaml
import dateparser
import pymongo
import gnupg
import json
from datetime import datetime

from kizano import getConfig, getLogger
from smartmetertx.api import MeterReader
from smartmetertx.utils import getMongoConnection

log = getLogger(__name__)
HOME = os.getenv('HOME', '')
SMTX_FROM   = dateparser.parse(os.environ.get('SMTX_FROM', 'day before yesterday'))
SMTX_TO     = dateparser.parse(os.environ.get('SMTX_TO', 'today'))

class Smtx2Mongo(object):
    '''
    SmartMeterTexas -> MongoDB
    Model object to take the records we get from https://smartmetertexas.com and insert them into
    a mongodb we control for data preservation and other analytics on our electric usage data we
    would want to undertake.
    '''

    def __init__(self):
        self.config = getConfig()
        self.mongo = getMongoConnection(self.config)
        self.db = self.mongo.get_database(self.config['mongo'].get('dbname', 'smartmetertx'))
        self.getSMTX()
        self.ensureIndexes()

    def close(self):
      if self.mongo:
        self.mongo.close()
        self.mongo = None

    def ensureIndexes(self):
        self.db.meterReads.create_index(
            [('reading', 1)],
            background=True
        )
        self.db.meterReads.create_index(
            [('datetime', 1)],
            background=True,
            unique=True
        )
        return self

    def getSMTX(self):
        log.info('Connecting to SmartMeterTX...')
        gpg = gnupg.GPG(gnupghome=os.path.join(os.getenv('HOME', '/home/markizano') , '.gnupg'), use_agent=True)
        self.smtx = MeterReader()
        smtx_user = self.config['smartmetertx']['user']
        smtx_pass = gpg.decrypt(self.config['smartmetertx']['pass']).data.decode('utf-8').strip()
        self.smtx.login(smtx_user, smtx_pass)
        log.info('Success! Getting meter reads...')
        return self.smtx

    def getDailyReads(self):
        # Get the meter reads and print the date in the format their API expects.
        reads = self.smtx.get_daily_read(self.config['smartmetertx']['esiid'], SMTX_FROM.strftime('%m/%d/%Y'), SMTX_TO.strftime('%m/%d/%Y'))
        if not reads:
            log.warn('Failed to get records from meterReads()')
        else:
            log.info('Acquired %d meter reads! Inserting into DB...' % len(reads['dailyData']))
        return reads

    def filterDailyReads(self, dailyData):
        results = []
        log.debug(json.dumps(dailyData, indent=2))
        for meterRead in dailyData:
            meterRead['datetime'] = datetime.strptime('%(date)s %(starttime)s' % meterRead, '%m/%d/%Y %H:%M%p')
            del meterRead['date'], meterRead['starttime']
            meterRead['startreading'] = float(meterRead['startreading'])
            meterRead['endreading'] = float(meterRead['endreading'])
            results.append(meterRead)
        return results

    def insertDailyData(self, dailyData):
        results = []
        log.info('Inserting %d reads into the DB.' % len(dailyData))
        try:
            insertResult = self.db.meterReads.insert_many(dailyData)
            log.debug(insertResult)
            results.append(insertResult)
        except pymongo.errors.BulkWriteError as e:
            errs = list(filter( lambda x: x['code'] != 11000, e.details['writeErrors'] ))
            if errs:
                raise errs
        log.info('Complete!')

def main():
    log.info('Gathering records from %s to %s' % ( SMTX_FROM.strftime('%F/%R'), SMTX_TO.strftime('%F/%R') ) )
    smtx2mongo = Smtx2Mongo()
    reads = smtx2mongo.getDailyReads()
    if not reads:
        log.error('Failed to read smartmetertexas API...')
        return 2

    dailyData = smtx2mongo.filterDailyReads(reads['dailyData'])
    if dailyData:
        smtx2mongo.insertDailyData(dailyData)
    else:
        log.warning('No records inserted!')
    smtx2mongo.close()
    return 0





import os
import json
import cherrypy

class SmartMeterController(object):
    '''
    Base Controller abstract.
    If a method is common among all controllers it goes here.
    If a method should be implemented by all controllers, it could be included
    here for all controllers to enjoy.
    '''

    def isValidJSONRequest(self, request, body):
        '''
        Collection of validations against the client request to this server.
        request: cherrypy.request object
        body: The result to attempt to parse as a string.

        400: bad request
        405: method not allowed
        406: not acceptable
        411: content-length required
        415: unsupported media type
        417: expectation failed
        422: Unprocessable entity

        Usage:

            isValid, errMesgs = self.isValidJSONRequest(request, payload)
            if not isValid:
                return { 'result': False, 'messages': errMesgs }
        '''
        result = True
        mesgs = []
        if request.get('REQUEST_METHOD', '') not in ['GET', 'POST', 'PUT']:
            result = False
            mesgs.append(ERRORS['request-method'])
            cherrypy.response.status = 405
        if request.get('REQUEST_URI', '').startswith('/api'):
            accept = request.get('HTTP_ACCEPT', '')
            if ('application/json' not in accept) and ('*/*' not in accept):
                result = False
                mesgs.append(ERRORS['accept-json'])
                cherrypy.response.status = 415
            if request['REQUEST_METHOD'] == 'POST':
                if 'application/json' not in request.get('CONTENT_TYPE', ''):
                    result = False
                    mesgs.append(ERRORS['not-json'])
                    cherrypy.response.status = 400
                cl = request.get('CONTENT_LENGTH', 0)
                if not cl or int(cl) < 1:
                    result = False
                    mesgs.append(ERRORS['content-length'])
                    cherrypy.response.status = 411
                if not body:
                    result = False
                    mesgs.append(ERRORS['empty-body'])
                    cherrypy.response.status = 411
                else:
                    try:
                        j = json.loads(body)
                    except ValueError as e:
                        result = False
                        mesgs.append(ERRORS['invalid-json'] % e)
                        cherrypy.response.status = 417
        return result, mesgs

    def returnError(self, err, mesgs):
        '''
        Return a [err] http response code and send the status message as an object.
        Usage:

          return self.returnError(500, ['no such file or directory'])
        '''
        kwargs = {}
        if not cherrypy.response.status or ( cherrypy.response.status >= 200 and cherrypy.response.status <= 299 ):
            cherrypy.response.status = int(err.get('status', 500))
        if isinstance(mesgs, str):
            mesgs = [mesgs]
        result = {
          'error': True,
          'mesgs': err.get('mesgs', mesgs),
          'value': None,
        }
        if 'DEBUG' in os.environ:
            kwargs['indent'] = 2
            kwargs['sort_keys'] = True
            if err.has_key('e'):
                result['exception'] = tb.format_exc(err['e'])
        return json.dumps(result, **kwargs) + "\n"

    def returnValue(self, result, value):
        '''
        Returns the value of the result you desire.
        result: Boolean to return as a result of the operation.
        value: object to return describing the result.
        @returns :string: JSON dump of a wrapped object of the input params.
        '''
        if not cherrypy.response.status:
            cherrypy.response.status = 200
        kwargs = {}
        if 'DEBUG' in os.environ:
            kwargs['indent'] = 2
            kwargs['sort_keys'] = True
        return json.dumps({
            'error': False,
            'status': 200,
            'value': value,
            'return': result,
        }, **kwargs) + "\n"




import os
import json
import requests

from pprint import pformat
from kizano import getLogger

# BEGIN: #StackOverflow
# @Source: https://stackoverflow.com/a/16630836/2769671
# These two lines enable debugging at httplib level (requests->urllib3->http.client)
# You will see the REQUEST, including HEADERS and DATA, and RESPONSE with HEADERS but without DATA.
# The only thing missing will be the response.body which is not logged.
if os.getenv('DEBUG', False):
    requests_log = getLogger("requests.urllib3", 10)
    requests_log.propagate = True
# END: #StackOverflow


class MeterReader:
    HOSTNAME = 'www.smartmetertexas.com'
    HOST = f'https://{HOSTNAME}'
    USER_AGENT = 'API Calls (python3; Linux x86_64) Track your own metrics with SmartMeterTX: https://github.com/markizano/smartmetertx'
    TIMEOUT = 30

    def __init__(self, timeout=10):
        self.log = getLogger(__name__)
        self.logged_in = False
        self.session = requests.Session()
        self.timeout = timeout
        self.session.headers['Authority'] = MeterReader.HOSTNAME
        self.session.headers['Origin'] = MeterReader.HOST
        self.session.headers['Accept'] = 'application/json, text/plain, */*'
        self.session.headers['Accept-Language'] = 'en-US,en;q=0.9'
        self.session.headers['Content-Type'] = 'application/json; charset=UTF-8'
        self.session.headers['dnt'] = '1'
        self.session.headers['sec-ch-ua'] = '".Not/A)Brand";v="99", "Google Chrome";v="103", "Chromium";v="103"'
        self.session.headers['sec-ch-ua-mobile'] = '?0'
        self.session.headers['sec-ch-ua-platform'] = 'Linux'
        self.session.headers['sec-fetch-dest'] = 'empty'
        self.session.headers['sec-fetch-mode'] = 'cors'
        self.session.headers['sec-fetch-site'] = 'same-origin'
        self.session.headers['User-Agent'] = MeterReader.USER_AGENT

    def api_call(self, url, json):
        '''
        Generic API call that can be made to the site for JSON results back.
        @param url :string: Where to send POST request.
        @param json :object: Data to send to the server.
        @return :object: JSON response back or ERROR
        '''
        self.log.debug(f'MeterReader.api_call(url={url}, json={json})')
        try:
            return self.session.post(
                url=url,
                json=json,
                timeout=self.timeout,
                verify=False
            )
        except Exception as ex:
            self.log.error(repr(ex))
            raise ex

    def login(self, username, password):
        '''
        Make API call to login and acquire a session token.
        @param username :string: Username or email used to Login to the webpage.
        @param password :string: Password used to login.
        @return :string: The login token that will be used going forward.
        '''
        creds = {
            "username": username,
            "password": password
        }
        url = f"{MeterReader.HOST}/commonapi/user/authenticate"
        r = self.api_call(url, json=creds)
        if r.status_code != 200:
            self.log.error("Login failed.")
            self.log.debug(pformat(r.headers.__dict__))
            self.log.debug(r.text)
            self.log.debug(self.session.cookies.__dict__)
            return False
        else:
            self.token = r.json()['token']
            self.log.info("Login successful!")
            self.log.debug(f"Got \x1b[33m{self.token}\x1b[0m as token.")
            self.session.headers["Authorization"] = f"Bearer {self.token}"
            self.logged_in = True
            return self.token

    def get_daily_read(self, esiid, start_date, end_date):
        if self.logged_in == False:
            self.log.error("You must login first.")
            return False

        json = {
            "esiid": esiid,
            "endDate": end_date,
            "startDate": start_date,
        }
        url = f"{MeterReader.HOST}/api/usage/daily"
        r = self.api_call(url, json=json)
        if r.status_code != 200 or "error" in r.text.lower():
            self.log.warning("Failed fetching daily read!")
            self.log.debug(r.text)
            self.log.debug(pformat(r.headers.__dict__))
            return False
        else:
            return r.json()




import os
import pymongo
import gnupg
import kizano
from urllib.parse import urlencode

log = kizano.getLogger(__name__)

def getMongoConnection(config: kizano.Config):
    # Establish connections to various sources and targets.
    log.info('Connecting to MongoDB...')
    gpg = gnupg.GPG(gnupghome=os.path.join(os.environ['HOME'], '.gnupg'), use_agent=True)
    mongo_user = config['mongo']['username']
    mongo_pass = gpg.decrypt(config['mongo']['password']).data.decode('utf-8')
    mongo_host = config['mongo']['host']
    mongo_opts = urlencode(config['mongo']['opts'])
    mongo_url = f'mongodb://{mongo_user}:{mongo_pass}@{mongo_host}/?{mongo_opts}'
    db = pymongo.MongoClient(mongo_url)
    log.info('Connected!')
    return db




#!/usr/bin/env python3

import os, io
import sys
from glob import glob
from pprint import pprint
from setuptools import setup

try:
    import yaml
except ImportError:
    os.system('pip3 install PyYAML')
    import yaml

sys.path.insert(0, os.path.abspath('.'))

try:
    PATCH = io.open('.build-id').read().strip()
except:
    try:
        pkginfo = yaml.safe_load(io.open('PKG-INFO').read())
        PATCH = pkginfo['Version'].split('.').pop()
    except Exception as e:
        PATCH = '0'

setup_opts = {
    'name'                : 'smartmetertx2mongo',
    # We change this default each time we tag a release.
    'version'             : f'1.1.{PATCH}',
    'description'         : 'Implementation of smartmetertx to save records to mongodb with config driven via YAML.',
    'author'              : 'Markizano Draconus',
    'author_email'        : 'markizano@markizano.net',
    'url'                 : 'https://markizano.net/',
    'license'             : 'GNU',
    'tests_require'       : ['nose', 'mock', 'coverage'],
    'install_requires'    : [
        'dateparser',
        'kizano',
        'pymongo',
        'requests',
        'cherrypy',
        'PyYAML',
        'python-gnupg',
        'jinja2',
    ],
    'package_dir'         : { 'smartmetertx': 'lib/smartmetertx' },
    'packages'            : [
      'smartmetertx',
    ],
    'scripts'             : glob('bin/*'),
    'test_suite'          : 'tests',
    'data_files'          : [
        ('share/smartmetertx', glob('ui/*')),
    ]
}

try:
    import argparse
    HAS_ARGPARSE = True
except:
    HAS_ARGPARSE = False

if not HAS_ARGPARSE: setup_opts['install_requires'].append('argparse')

# I botch this too many times.
if sys.argv[1] == 'test':
    sys.argv[1] = 'nosetests'

if 'DEBUG' in os.environ: pprint(setup_opts)

setup(**setup_opts)

